import os
from signal import SIGTERM

import pytest

from .mock_objects import mock_get_all_workers_on_namespace
from skidward.__main__ import CLICommands, _detect_new_workers
from skidward.models import db, Worker, User, Task
import skidward.worker_detector as wd


def test_ensure_test_database(app):
    assert "skidwardDB_TEST" in app.config["SQLALCHEMY_DATABASE_URI"]


def test_detect_new_workers(monkeypatch, app):
    monkeypatch.setattr(
        wd, "get_all_workers_on_namespace", mock_get_all_workers_on_namespace
    )
    with app.app_context():
        test_ensure_test_database(app)

        expected_result = ["real_name", "other_name"]
        actual_result = _detect_new_workers()

        assert len(expected_result) == len(actual_result)
        assert expected_result.pop() in actual_result
        assert expected_result.pop() in actual_result


def test_workers_publishable_in_db(monkeypatch, real_namespace, app):
    monkeypatch.setattr(
        wd, "get_all_workers_on_namespace", mock_get_all_workers_on_namespace
    )
    with app.app_context():
        test_ensure_test_database(app)

        workers_on_namespace = wd.get_all_workers_on_namespace(real_namespace)
        registered_workers = [w.name for w in db.session.query(Worker).all()]
        unregistered_workers = list(set(workers_on_namespace) - set(registered_workers))

        # Verify workers found different than ones in DB
        assert registered_workers is not None
        assert len(workers_on_namespace) > 0
        assert len(unregistered_workers) > 0

        CLICommands.publish_workers()

        new_registered_workers = [w.name for w in db.session.query(Worker).all()]
        new_unregistered_workers = list(
            set(workers_on_namespace) - set(new_registered_workers)
        )

        # Verify new workers registered in DB
        assert new_registered_workers is not None
        assert len(new_registered_workers) > 0
        assert len(new_unregistered_workers) == 0


def test_scheduler_spawns_new_processes(app):
    with app.app_context():
        worker = Worker(name="worker", description="A worker.")
        db.session.add(worker)
        db.session.commit()

        context = {"LOGGING_CHANNEL": "ch:channel"}
        task_one = Task(
            name="first_task",
            worker=worker,
            worker_id=worker.id,
            context=context,
            cron_string="",
        )
        task_two = Task(
            name="second_task",
            worker=worker,
            worker_id=worker.id,
            context=context,
            cron_string="",
        )

        db.session.add(task_one)
        db.session.add(task_two)
        db.session.commit()

    pid_one = CLICommands.start_scheduler(True)
    pid_two = CLICommands.start_scheduler(True)

    assert pid_one is not pid_two

    # Cleanup running processes
    os.kill(pid_one, SIGTERM)
    os.kill(pid_two, SIGTERM)


@pytest.mark.parametrize(
    ("email"), ("dink@dink.com", "admin@skidward.com", "admin", "", 123)
)
def test_create_admin(monkeypatch, app, email):
    monkeypatch.setattr("getpass.getpass", lambda x: "password")
    with app.app_context():
        test_ensure_test_database(app)

        try:
            assert len(db.session.query(User).all()) == 0
            CLICommands.create_admin(email)
            assert db.session.query(User).all() is not None
        except Exception:
            pytest.fail("Admin could not be created.")



from flask import request, redirect, url_for, flash
from flask_admin.menu import MenuLink
from wtforms.fields import PasswordField
from flask_admin.form import fields
from flask_admin.model.template import EndpointLinkRowAction
from flask_admin.contrib.sqla import ModelView
from flask_admin import AdminIndexView, expose
from flask_security import login_required, current_user, utils, logout_user, login_user

from skidward.models import db, Task
from skidward.backend import RedisBackend
from skidward.web.forms import TaskContextForm


class SkidwardModelView(ModelView):
    def is_accessible(self):
        return current_user.is_authenticated

    def inaccessible_callback(self, name, **kwargs):
        return redirect(url_for("index.login"))


class UserAdmin(SkidwardModelView):
    column_exclude_list = ["password"]
    form_excluded_columns = ("password",)

    def is_accessible(self):
        return current_user.is_authenticated and current_user.has_role("admin")

    def scaffold_form(self):
        form_class = super(UserAdmin, self).scaffold_form()
        form_class.password2 = PasswordField("New Password")
        return form_class

    def on_model_change(self, form, model, is_created):
        if len(model.password2):
            model.password = utils.hash_password(model.password2)


class RoleAdmin(SkidwardModelView):
    def is_accessible(self):
        return current_user.is_authenticated and current_user.has_role("admin")


class WorkerView(SkidwardModelView):
    can_create = False


class TaskView(SkidwardModelView):

    column_extra_row_actions = [
        EndpointLinkRowAction(
            icon_class="glyphicon glyphicon-cog",
            endpoint="add_context.add_context",
            title="Configure",
            id_arg="task_id",
        ),
        EndpointLinkRowAction(
            icon_class="glyphicon glyphicon-play-circle",
            endpoint="add_context.add_context",
            title="Configured_Run",
            id_arg="task_id",
            url_args=dict(temp_run=True),
        ),
        EndpointLinkRowAction(
            icon_class="glyphicon glyphicon-play",
            endpoint="run_task.run_task",
            title="Quick_Run",
            id_arg="task_id",
        ),
    ]

    form_overrides = {"context": fields.JSONField}
    form_excluded_columns = ("context",)
    form_create_rules = ("name", "cron_string", "worker")


class JobView(SkidwardModelView):

    column_list = ["task", "ran_at", "state"]
    column_labels = dict(task="Jobs", state="Status")


class SkidwardView(AdminIndexView):
    def is_accessible(self):
        return current_user.is_authenticated

    def inaccessible_callback(self, name, **kwargs):
        return redirect(url_for("index.login"))

    def is_visible(self):
        return False

    @login_required
    @expose("/")
    def index(self):
        return self.render("admin/index.html")

    @expose("/login")
    def login(self):
        login_user(current_user)
        return redirect(url_for("admin.index"))

    @expose("/logout")
    def logout(self):
        logout_user()
        return redirect(url_for("index.login"))


class RunView(SkidwardView):
    @expose("/run/<task_id>")
    @login_required
    def run_task(self, task_id):
        task = Task.query.get(task_id)
        context = request.args.get("context")

        if not context and not task.context:
            flash("Cannot Run Task without Configuration.", category="error")
            return redirect(url_for("add_context.add_context", task_id=task_id))

        redis_client = RedisBackend()
        if context:
            redis_client.hset("MANUAL_RUN", task_id, context)
            return self.render("task/run_task.html", name=task.name, context=context)

        else:
            redis_client.lpush("MANUAL_RUN", task_id)
            return self.render(
                "task/run_task.html", name=task.name, context=task.context
            )


class TaskConfigure(SkidwardView):
    @expose("/configure/<task_id>", methods=["GET", "POST"])
    @login_required
    def add_context(self, task_id):
        form = TaskContextForm()
        task = Task.query.get(task_id)

        if form.validate_on_submit():
            dict_of_fieldlist = request.form.to_dict(flat=False)
            keys = dict_of_fieldlist["key"]
            values = dict_of_fieldlist["value"]
            context = dict(zip(keys, values))

            if request.args.get("temp_run"):
                return redirect(
                    url_for("run_task.run_task", task_id=task.id, context=context)
                )

            else:
                task.context = context
                db.session.commit()
                return redirect(url_for("add_context.index"))

        return self.render("task/configure.html", form=form, name=task.name)


class LoginMenuLink(MenuLink):
    def is_accessible(self):
        return not current_user.is_authenticated


class LogoutMenuLink(MenuLink):
    def is_accessible(self):
        return current_user.is_authenticated



import os

import pytest
from flask_security import Security, SQLAlchemyUserDatastore

from skidward.app import app as application
from skidward.models import db, User, Role, Task, Worker
from skidward.backend import RedisBackend
from tests.mock_objects import MockNamespaceModuleManager


@pytest.fixture(autouse=True, scope="session")
def set_testing_state():
    os.environ["WEB_TESTING"] = "True"
    yield
    os.environ["WEB_TESTING"] = "False"


@pytest.fixture(scope="function")
def app(request):
    def cleanup_db():
        with application.app_context():
            db.session.rollback()
            db.drop_all()

    def set_test_db_in_application():
        with application.app_context():
            db_uri = application.config["SQLALCHEMY_DATABASE_URI"].split("/")
            db_uri[-1] = "skidwardDB_TEST"
            application.config["SQLALCHEMY_DATABASE_URI"] = "/".join(db_uri)

    with application.app_context():
        set_test_db_in_application()
        db.init_app(application)
        Security(
            application,
            SQLAlchemyUserDatastore(db, User, Role),
            register_blueprint=False,
        )
        db.create_all()
        request.addfinalizer(cleanup_db)

        return application


@pytest.fixture()
def init_database():
    with application.app_context():
        db.create_all()
        worker_one = Worker(name="real_name")
        db.session.add(worker_one)
        db.session.commit()

        task_one = Task(
            name="task_1",
            worker=worker_one,
            context={"setting_1": "value_1"},
            cron_string="*/2 * * * *",
        )
        db.session.add(task_one)
        db.session.commit()
        yield db
        db.session.remove()
        db.drop_all()


@pytest.fixture
def real_namespace() -> str:
    return "skidward.real.name.space"


@pytest.fixture
def mock_namespace_module_manager() -> MockNamespaceModuleManager:
    return MockNamespaceModuleManager("skidward.real.name.space")


@pytest.fixture()
def backend(monkeypatch):
    monkeypatch.setenv("WEB_TESTING", "True")
    dummy_backend = RedisBackend()
    yield dummy_backend
    dummy_backend.erase()



from pkg_resources import EntryPoint

from ..mock_objects import mock_create_namespace_module_manager
from skidward import worker_detector as wd


def test_get_default_namespace():
    assert isinstance(wd.__NAMESPACE, str)
    assert wd.__NAMESPACE == "skidward.workers"


def test_get_workers_without_providing_namespace():
    all_workers = wd.get_all_workers_on_namespace()

    assert isinstance(all_workers, list)


def test_get_workers_on_fake_provided_namespace():
    all_workers = wd.get_all_workers_on_namespace("skidward.fake.name.space")

    assert isinstance(all_workers, list)
    assert len(all_workers) == 0


def test_get_workers_on_real_provided_namespace(monkeypatch):
    monkeypatch.setattr(
        wd, "create_namespace_module_manager", mock_create_namespace_module_manager
    )
    all_workers = wd.get_all_workers_on_namespace("skidward.real.name.space")

    assert isinstance(all_workers, list)
    assert len(all_workers) > 0
    assert isinstance(all_workers.pop(), str)


def test_load_fake_worker_on_default_namespace(capsys):
    wd.load_worker_on_namespace("fake_name")
    captured = capsys.readouterr()

    assert "No worker could be loaded" in captured.out
    assert "fake_name" in captured.out


def test_load_real_worker_on_fake_namespace(real_namespace, monkeypatch, capsys):
    monkeypatch.setattr(
        wd, "create_namespace_module_manager", mock_create_namespace_module_manager
    )
    all_workers = wd.get_all_workers_on_namespace(real_namespace)
    assert len(all_workers) > 0

    first_worker = all_workers.pop()
    assert isinstance(first_worker, str)

    wd.load_worker_on_namespace(first_worker, "skidward.fake.name.space")
    captured = capsys.readouterr()

    assert "No worker could be loaded" in captured.out
    assert first_worker in captured.out


def test_load_real_worker_on_real_namespace(real_namespace, monkeypatch):
    monkeypatch.setattr(
        wd, "create_namespace_module_manager", mock_create_namespace_module_manager
    )
    all_workers = wd.get_all_workers_on_namespace(real_namespace)
    assert len(all_workers) > 0

    first_worker = all_workers.pop()
    load_result = wd.load_worker_on_namespace(first_worker, real_namespace)

    assert load_result is not None
    assert isinstance(load_result, EntryPoint)
    assert load_result.module_name == first_worker



import os
import logging
from functools import partialmethod

from redis import StrictRedis


logger = logging.getLogger(__name__)


class RedisProxy:
    def __init__(self, backend):
        self.backend = backend

    def _decode_response(self, values):
        if isinstance(values, bytes):
            return values.decode("utf-8")

        if isinstance(values, list):
            new_values = []
            for element in values:
                if isinstance(element, bytes):
                    element = element.decode("utf-8")
                new_values.append(element)
            return new_values

        return values

    def proxy(self, method, *args, **kwargs):
        errors = {}

        backend_name = self.backend.__class__.__name__
        try:
            logger.debug(
                "trying to {} {} using {}".format(method, kwargs, backend_name)
            )
            bound_method = getattr(self.backend, method)
            result = bound_method(*args, **kwargs)

            return self._decode_response(result)
        except Exception as e:
            if os.getenv("DEBUG"):
                raise
            errors[backend_name] = str(e)

        raise Exception(
            "No backend managed to {} for the provided request: {}".format(
                method, errors
            )
        )

    lpush = partialmethod(proxy, "lpush")
    lrange = partialmethod(proxy, "lrange")
    erase = partialmethod(proxy, "erase")
    set = partialmethod(proxy, "set")
    get = partialmethod(proxy, "get")
    rpop = partialmethod(proxy, "rpop")
    exists = partialmethod(proxy, "exists")
    hset = partialmethod(proxy, "hset")


class RedisDummyBackend:
    def __init__(self, **kwargs):
        self.redis_lists = {"set": {}}

    def is_task_new(self, name):
        return name not in self.redis_lists and name not in self.redis_lists["set"]

    def lpush(self, name, elements):
        if not isinstance(elements, list):
            elements = [elements]

        elements = [str(element).encode("utf-8") for element in elements]

        if not name in self.redis_lists:
            self.redis_lists[name] = []

        self.redis_lists[name] = elements + self.redis_lists[name]

    def lrange(self, name, start, end):
        if self.is_task_new(name):
            self.redis_lists[name] = []

        if end == -1:
            return self.redis_lists[name][start:]
        else:
            end = end + 1

        result = self.redis_lists[name][start:end]

        return result

    def erase(self):
        self.redis_lists = {}

    def set(self, name, value, ex=None, nx=None):
        if nx:
            if self.is_task_new(name):
                self.redis_lists["set"][name] = str(value).encode("utf-8")
        else:
            self.redis_lists["set"][name] = str(value).encode("utf-8")

    def get(self, name):
        if not self.is_task_new(name):
            return self.redis_lists["set"][name]

    def rpop(self, name):
        if not self.is_task_new(name):
            if self.redis_lists[name]:
                return self.redis_lists[name].pop(-1)

    def exists(self, name):
        return not self.is_task_new(name)

    def hset(self, name, key, value):
        if not name in self.redis_lists:
            self.redis_lists[name] = []

        self.redis_lists[name].append({"task": key, "context": value})


def RedisBackend(**kwargs):
    def _is_testing():
        return os.getenv("WEB_TESTING") is True or os.getenv("WEB_TESTING") in ["True", "true", "TRUE"]

    if _is_testing():
        return RedisProxy(RedisDummyBackend(**kwargs))

    return RedisProxy(StrictRedis(**kwargs))




import pytest

from skidward.app import app
from skidward.models import db


def test_it_creates_app():
    assert app


def test_db_initialized_with_app():
    assert db
    with app.app_context():
        assert db.get_app() == app


@pytest.mark.parametrize(
    "config_key",
    (
        "DEBUG",
        "FLASK_ADMIN_SWATCH",
        "SQLALCHEMY_DATABASE_URI",
        "SECRET_KEY",
        "SECURITY_PASSWORD_SALT",
        "SECURITY_PASSWORD_HASH",
        "SQLALCHEMY_TRACK_MODIFICATIONS",
    ),
)
def test_app_is_configured(config_key):
    with app.app_context():
        try:
            assert app.config[config_key] is not None
        except KeyError:
            pytest.fail(f"Key {config_key} not found in the app config.")



from wtforms import StringField, validators
from flask_wtf import FlaskForm


class TaskContextForm(FlaskForm):
    key = StringField("Key", [validators.DataRequired()])
    value = StringField("Value", [validators.DataRequired()])



import os

from flask import Flask
from flask_security import SQLAlchemyUserDatastore, Security

from skidward.models import db, User, Role


app = Flask(__name__, template_folder="web/templates")

app.debug = os.getenv("FLASK_DEBUG")
# Setting any FLASK_ADMIN_SWATCH(Theme Template)
app.config["FLASK_ADMIN_SWATCH"] = os.getenv("FLASK_ADMIN_SWATCH")

# Setting up Postgres Database
app.config["SQLALCHEMY_DATABASE_URI"] = os.getenv("SQLALCHEMY_DATABASE_URI")

# Setting up a secret key
app.config["SECRET_KEY"] = os.getenv("SECRET_KEY")
app.config["SECURITY_PASSWORD_SALT"] = os.getenv("SECURITY_PASSWORD_SALT")
app.config["SECURITY_PASSWORD_HASH"] = os.getenv("SECURITY_PASSWORD_HASH")
app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = os.getenv(
    "SQLALCHEMY_TRACK_MODIFICATIONS"
)


db.init_app(app)
security = Security(app, SQLAlchemyUserDatastore(db, User, Role))

with app.app_context():
    db.create_all()



import os
import time
from itertools import chain
from datetime import datetime as dt
from multiprocessing import Process

from cronex import CronExpression

from skidward.backend import RedisBackend
from skidward.models import Task
from skidward.task_runner import TaskRunner


REDIS_ELECTION_TTL = 5
REDIS_MANUAL_CHANNEL = "MANUAL_RUN"
WAKEUP_TIMER = int(os.getenv("WAKEUP_TIMER"))

process_id = os.getpid()


class Elector:
    def __init__(self, redis_client):
        self.redis_client = redis_client

    def get_tasks(self):
        for task in chain(self.get_scheduled_tasks(), self.get_manual_tasks()):
            yield task

    def get_scheduled_tasks(self):
        for task in self.get_elected_tasks(
            Task.query.filter(Task.cron_string != None, Task.cron_string != "").all()
        ):
            yield task

    def get_manual_tasks(self):
        task_id = self.redis_client.rpop(REDIS_MANUAL_CHANNEL)
        while task_id:
            yield Task.query.get(task_id)
            task_id = self.redis_client.rpop(REDIS_MANUAL_CHANNEL)

    def get_elected_tasks(self, tasks):
        return [task for task in tasks if self.is_runner_elected_for_task(task)]

    def is_runner_elected_for_task(self, task):
        date_tuple = dt.utcnow().timetuple()[:5]
        date_mask = "-".join(map(str, date_tuple))
        unique_key = f"{task.id}-{date_mask}"

        cron_expression = CronExpression(task.cron_string)
        if cron_expression.check_trigger(date_tuple):
            if not self.redis_client.exists(unique_key):
                self.redis_client.set(
                    unique_key, process_id, ex=REDIS_ELECTION_TTL, nx=True
                )
                return self.redis_client.get(unique_key) == str(process_id)

        return False


class SchedulerRunner:
    def __init__(self):
        self.elector = Elector(RedisBackend())

    def run(self):
        while True:
            for task in self.elector.get_tasks():
                runner = TaskRunner(task)
                Process(target=runner.start).start()

            time.sleep(WAKEUP_TIMER)


def start():
    SchedulerRunner().run()



import logging
import datetime
from unittest import mock

from skidward.app import app
from skidward.models import Job, JobStatus, Task, Worker
from skidward.task_runner import TaskRunner


def test_new_job_is_created(init_database, backend, monkeypatch):
    monkeypatch.setattr(
        TaskRunner,
        "_set_logging_config",
        (lambda x: logging.basicConfig(handlers=logging.NullHandler())),
    )
    with app.app_context():
        worker = Worker.query.all()[0]
        context = {"LOGGING_CHANNEL": "ch:channel"}
        task = Task(name="task_id", worker=worker, context=context, cron_string="")
        init_database.session.add(task)
        init_database.session.commit()

        TaskRunner(task)._create_new_job()

        assert init_database.session.query(Job).one().task.id == task.id


@mock.patch("skidward.worker_detector.create_namespace_module_manager")
def test_full_process_calling_run(mock_create, init_database, monkeypatch):
    monkeypatch.setattr(
        TaskRunner,
        "_set_logging_config",
        (lambda x: logging.basicConfig(handlers=logging.NullHandler())),
    )
    with app.app_context():
        worker = Worker.query.all()[0]
        context = {"LOGGING_CHANNEL": "ch:channel"}
        task = Task(name="task_id", worker=worker, context=context, cron_string="")
        init_database.session.add(task)
        init_database.session.commit()

        TaskRunner(task).start()
        job = init_database.session.query(Job).one()

        mock_create.return_value.load_module.assert_called_once_with("real_name")
        mock_create.return_value.load_module.return_value.start.assert_called_once_with(
            {"LOGGING_CHANNEL": "ch:channel"}
        )
        assert job.id == 1
        assert job.task_id == 2
        assert job.state == JobStatus.SUCCESS
        assert isinstance(job.ran_at, datetime.datetime)



import pkg_resources as pkg
from typing import Dict, List

from skidward.worker_detector import NamespaceModuleManager


class MockEntryPoint(pkg.EntryPoint):
    def __init__(self, name, module_name):
        self.name = name
        self.module_name = module_name
        self.attrs = []
        self.extras = []
        self.dist = []

    def load(self, *args, **kwargs) -> pkg.EntryPoint:
        return MockEntryPoint(self.name, self.module_name)

    def start(self, context):
        return True


class MockNamespaceModuleManager(NamespaceModuleManager):
    def __init__(self, namespace):
        super().__init__(namespace)

    def _get_module_entry_points_on_namespace(self) -> Dict[str, pkg.EntryPoint]:
        if self.namespace == "skidward.real.name.space":
            return {
                "real_name": MockEntryPoint("real-name", "real_name"),
                "other_name": MockEntryPoint("other-name", "other_name"),
            }

        return super()._get_module_entry_points_on_namespace()


def mock_create_namespace_module_manager(namespace: str) -> MockNamespaceModuleManager:
    return MockNamespaceModuleManager(namespace)


def mock_get_all_workers_on_namespace(namespace: str = None) -> List[str]:
    namespace = "skidward.real.name.space"
    nmm = mock_create_namespace_module_manager(namespace)
    return nmm.get_all_available_module_names()



from datetime import datetime as dt
from unittest import mock

import pytest

from skidward.models import Task, Worker
from skidward.scheduler import scheduler


class TestElector:
    @mock.patch("skidward.scheduler.scheduler.CronExpression.check_trigger")
    def test_when_a_task_is_runner_elected_for_task(
        self, mock_cron, init_database, backend
    ):
        mock_cron.return_value = True

        elector = scheduler.Elector(backend)
        assert elector.is_runner_elected_for_task(Task.query.one())

    @mock.patch("skidward.scheduler.scheduler.CronExpression.check_trigger")
    def test_when_a_task_is_not_ready_to_run(self, mock_cron, init_database, backend):
        mock_cron.return_value = False

        elector = scheduler.Elector(backend)
        assert not elector.is_runner_elected_for_task(Task.query.one())

    @mock.patch("skidward.scheduler.scheduler.CronExpression.check_trigger")
    def test_task_is_not_ready_when_already_exists_in_redis(
        self, mock_cron, init_database, backend
    ):
        mock_cron.return_value = True

        date_mask = "-".join(map(str, dt.utcnow().timetuple()[:5]))
        task = Task.query.one()
        unique_key = f"{task.id}-{date_mask}"
        backend.set(unique_key, "host_1", ex=5, nx=True)

        elector = scheduler.Elector(backend)
        assert not elector.is_runner_elected_for_task(task)

    @mock.patch("skidward.backend.RedisDummyBackend.exists")
    @mock.patch("skidward.scheduler.scheduler.CronExpression.check_trigger")
    def test_task_is_not_ready_if_another_scheduler_holds_it(
        self, mock_cron, mock_redis_exists, init_database, backend
    ):
        mock_cron.return_value = True
        mock_redis_exists.return_value = False

        date_mask = "-".join(map(str, dt.utcnow().timetuple()[:5]))
        task = Task.query.one()
        unique_key = f"{task.id}-{date_mask}"
        backend.set(unique_key, "host_1", ex=5, nx=True)

        elector = scheduler.Elector(backend)
        assert not elector.is_runner_elected_for_task(Task.query.one())

    @pytest.mark.parametrize("length, ready", [(1, True), (0, False)])
    @mock.patch("skidward.scheduler.scheduler.Elector.is_runner_elected_for_task")
    def test_get_list_of_eligible_tasks(
        self, mock_ready, length, ready, init_database, backend
    ):
        mock_ready.return_value = ready

        elector = scheduler.Elector(backend)
        eligible_tasks = elector.get_elected_tasks(Task.query.all())
        assert len(eligible_tasks) == length

    def test_get_manual_task_if_available_in_redis_channel(
        self, backend, init_database
    ):
        task_id = Task.query.one().id
        backend.lpush("MANUAL_RUN", task_id)

        elector = scheduler.Elector(backend)
        yield_task = next(elector.get_manual_tasks())
        assert isinstance(yield_task, Task)
        assert yield_task.id == task_id

    def test_do_not_get_manual_task_if_not_available_in_redis_channel(
        self, backend, init_database
    ):
        task = Task.query.one()

        elector = scheduler.Elector(backend)
        with pytest.raises(StopIteration):
            next(elector.get_manual_tasks())

    @mock.patch("skidward.scheduler.scheduler.Elector.get_elected_tasks")
    def test_get_scheduled_task_if_eligible(self, mock_elected, init_database, backend):
        task = Task.query.one()
        mock_elected.return_value = [task]

        elector = scheduler.Elector(backend)
        yield_task = next(elector.get_scheduled_tasks())
        assert isinstance(yield_task, Task)
        assert yield_task.id == task.id

    @mock.patch("skidward.scheduler.scheduler.Elector.get_elected_tasks")
    def test_do_not_get_scheduled_task_if_not_eligible(
        self, mock_elected, backend, init_database
    ):
        mock_elected.return_value = []

        elector = scheduler.Elector(backend)
        with pytest.raises(StopIteration):
            next(elector.get_scheduled_tasks())

    @mock.patch("skidward.scheduler.scheduler.Elector.get_scheduled_tasks")
    @mock.patch("skidward.scheduler.scheduler.Elector.get_manual_tasks")
    def test_get_manual_and_schedule_tasks(
        self, mock_schedule, mock_manual, backend, init_database
    ):
        task = Task.query.one()
        worker = Worker.query.one()

        task_two = Task(
            name="task_2",
            worker_id=worker.id,
            context={"setting_1": "value_1"},
            cron_string="*/2 * * * *",
        )
        init_database.session.add(task_two)
        init_database.session.commit()

        mock_schedule.return_value = iter([task])
        mock_manual.return_value = iter([task_two])

        elector = scheduler.Elector(backend)
        tasks = elector.get_tasks()
        yield_task = next(tasks)
        assert isinstance(yield_task, Task)
        assert yield_task.id == task_two.id

        yield_task = next(tasks)
        assert isinstance(yield_task, Task)
        assert yield_task.id == task.id

    @mock.patch("skidward.scheduler.scheduler.Elector.get_scheduled_tasks")
    @mock.patch("skidward.scheduler.scheduler.Elector.get_manual_tasks")
    def test_get_manual_task_but_not_scheduled_as_not_available(
        self, mock_schedule, mock_manual, backend, init_database
    ):
        worker = Worker.query.one()
        task_two = Task(
            name="task_2",
            worker_id=worker.id,
            context={"setting_1": "value_1"},
            cron_string="*/2 * * * *",
        )
        init_database.session.add(task_two)
        init_database.session.commit()

        mock_schedule.return_value = None
        mock_manual.return_value = iter([task_two])

        elector = scheduler.Elector(backend)
        tasks = elector.get_tasks()
        yield_task = next(tasks)
        assert isinstance(yield_task, Task)
        assert yield_task.id == task_two.id

        with pytest.raises(TypeError):
            next(tasks)



import pytest
import os

from skidward.web import app
from skidward.models import db, User, Worker, Task


def create_app():
    app.config["TESTING"] = os.getenv("WEB_TESTING")
    app.config["WTF_CSRF_ENABLED"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///:memory:"
    app.config["SECRET_KEY"] = os.getenv("SECRET_KEY")
    return app


@pytest.fixture()
def test_client():
    flask_app = create_app()
    client = flask_app.test_client()
    ctx = flask_app.app_context()
    ctx.push()
    yield client
    ctx.pop()


@pytest.fixture()
def init_database():
    db.create_all()
    user1 = User(email="test@test.com", username="testuser", password="123")
    db.session.add(user1)
    worker1 = Worker(name="Test_Worker")
    db.session.add(worker1)
    task1 = Task(name="Test_Task", worker=worker1, context={"test_key": "test_value"})
    db.session.add(task1)
    db.session.commit()
    yield db
    db.drop_all()



from unittest import mock

from flask_login import current_user

from skidward.models import User, Worker, Task


def test_it_loads_the_app(test_client):
    response = test_client.get("/login")
    assert response.status_code == 200


def test_it_requires_login_to_access_home_page(test_client):
    response = test_client.get("/", follow_redirects=True)
    assert b"Login" in response.data


def test_it_logs_in_correct_user(test_client, init_database):
    with test_client:
        response = test_client.post(
            "/login",
            data=dict(email="test@test.com", password="123"),
            follow_redirects=True,
        )
        assert current_user.email == "test@test.com"


def test_it_creates_new_user_in_the_database(test_client, init_database):
    user = User.query.filter_by(email="test@test.com").first()
    assert user is not None
    assert user.email == "test@test.com"
    assert user.username == "testuser"


def test_it_redirects_to_admin_interface_on_login(test_client, init_database):
    response = test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    assert b"Logged in as" in response.data


def test_it_logs_out_correctly_and_redirects_to_login(test_client, init_database):
    response = test_client.get("/logout", follow_redirects=True)
    assert response.status_code == 200
    assert not current_user
    assert b"Login" in response.data


def test_configure_asks_for_context_and_save_in_db_correctly(
    test_client, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    response = test_client.get("/admin/task/configure/1", follow_redirects=True)
    assert b"Add Configuration" in response.data
    test_client.post(
        "/admin/task/configure/1",
        data=dict(key="new_key", value="new_value"),
        follow_redirects=True,
    )
    task = Task.query.get(1)
    assert task.context == {"new_key": "new_value"}
    assert type(task.context) == dict


def test_configured_run_takes_new_context_but_does_not_save_in_db(
    backend, test_client, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    response = test_client.post(
        "/admin/task/configure/1?temp_run=True",
        data=dict(key="tmp_key", value="tmp_value"),
        follow_redirects=True,
    )
    assert b"RUNNING TASK" in response.data
    task = Task.query.get(1)
    assert task.context == {"test_key": "test_value"}


@mock.patch("skidward.backend.RedisDummyBackend")
def test_quick_run_pushes_task_id_to_redis_and_redirects_to_run_task(
    mock_redis, test_client, backend, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    mock_redis.return_value = backend
    response = test_client.get("/admin/task/run/1", follow_redirects=True)
    redis_tasks = backend.lrange("MANUAL_RUN", 0, -1)

    assert b"RUNNING TASK" in response.data
    assert "1" in redis_tasks


@mock.patch("skidward.backend.RedisDummyBackend")
def test_configured_run_pushes_dict_of_task_with_context_to_redis(
    mock_redis, test_client, backend, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    mock_redis.return_value = backend
    test_client.post(
        "admin/task/configure/1?temp_run=True",
        data=dict(key="tmp_key", value="tmp_value"),
        follow_redirects=True,
    )
    redis_tasks = backend.lrange("MANUAL_RUN", 0, -1)
    for t in redis_tasks:
        assert t["task"] == "1"
        assert t["context"] == "{'tmp_key': 'tmp_value'}"


def test_manual_run_redirects_to_configure_if_no_context_is_present(
    test_client, init_database
):
    test_client.post(
        "/login",
        data=dict(email="test@test.com", password="123"),
        follow_redirects=True,
    )
    worker = Worker.query.first()
    new_task = Task(name="New_Task", worker=worker)
    init_database.session.add(new_task)
    init_database.session.commit()
    task = Task.query.filter_by(name="New_Task").first()
    assert task is not None
    assert task.id == 2
    assert not task.context
    response = test_client.get("/admin/task/run/2", follow_redirects=True)
    assert b"Add Configuration" in response.data



import datetime
import logging

from redis_log_handler.RedisLogHandler import RedisLogHandler

from skidward import worker_detector
from skidward.models import db, Job, JobStatus


class TaskRunner:
    def __init__(self, task):
        self.task = task
        self.context = self.task.context
        self.worker_name = self.task.worker.name

    def _create_new_job(self):
        job = Job(
            state=JobStatus.RUNNING,
            ran_at=datetime.datetime.utcnow(),
            task_id=self.task.id,
        )
        db.session.add(job)
        db.session.commit()

        logging.info("Job id:{} created".format(job.id))

        return job

    def _set_logging_config(self):
        log_handler = RedisLogHandler(self.context.get("LOGGING_CHANNEL"))
        logger = logging.getLogger()
        logger.addHandler(log_handler)
        logger.setLevel(logging.INFO)

        logging.info("Logging {}".format(self.worker_name))

    def start(self):
        self._set_logging_config()

        job = self._create_new_job()
        worker_module = worker_detector.load_worker_on_namespace(self.worker_name)

        logging.info("{} is running".format(self.worker_name))

        try:
            worker_module.start(self.context)
            status = JobStatus.SUCCESS
        except:
            status = JobStatus.FAIL

        logging.info("Status : {}".format(status))

        job.state = status
        db.session.commit()



import os
import getpass
import multiprocessing

import fire
import click
from flask_security import utils
from werkzeug.local import LocalProxy

from skidward import worker_detector
from skidward.app import app
from skidward.scheduler import scheduler
from skidward.models import db, Worker


def _initialize_database():
    db.init_app(app)
    db.create_all()
    print("Database initialized.")


def _migrate_db_if_needed():
    # We currently do not support migrations, not sure if we want/need this
    # If we use Flask-migrate:

    # from flask-migrate import Migrate, migrate
    # migrations = Migrate(app, db)
    # migrate()
    print("Migrations not yet implemented.")
    pass


def _detect_new_workers() -> list:
    def _get_registered_workers() -> list:
        return [w.name for w in Worker.query.all()]

    unregistered_workers = []
    registered_workers = _get_registered_workers()
    detected_workers = worker_detector.get_all_workers_on_namespace()

    for dw in detected_workers:
        if dw not in registered_workers:
            unregistered_workers.append(dw)

    return unregistered_workers


def _insert_new_workers_in_database():
    new_workers = [Worker(name=worker) for worker in _detect_new_workers()]
    db.session.bulk_save_objects(new_workers)
    db.session.commit()


def _start_new_scheduler(is_background_process):
    if is_background_process:
        process = multiprocessing.Process(target=scheduler.start)
        process.start()
        return process.pid

    print(os.getpid())
    return scheduler.start()


def _create_admin_user(email):
    email = str(email)
    data_store = LocalProxy(lambda: app.extensions["security"].datastore)

    def _get_user_by_email():
        return data_store.find_user(email=email)

    def _set_admin_role_for_user():
        data_store.add_role_to_user(email, "admin")

    def _prompt_for_password():
        return utils.hash_password(getpass.getpass("Please enter your password: "))

    def _create_user():
        hashed_password = _prompt_for_password()
        username = email.split("@")[0]

        new_user = data_store.create_user(
            email=email, username=username, password=hashed_password
        )
        data_store.add_role_to_user(email, "admin")

        return new_user

    data_store.find_or_create_role(name="admin", description="Administrator")
    data_store.find_or_create_role(name="end-user", description="End user")

    existing_user = _get_user_by_email()
    if existing_user:
        if click.confirm("User exists. Upgrade to admin?"):
            _set_admin_role_for_user()
            print("User upgraded to admin.")

        return existing_user

    new_user = _create_user()
    _set_admin_role_for_user()
    db.session.commit()

    print("User created and set to admin.")

    return new_user


class CLICommands(object):
    @staticmethod
    def initdb():
        _initialize_database()

    @staticmethod
    def migrate():
        _migrate_db_if_needed()

    @staticmethod
    def publish_workers():
        _insert_new_workers_in_database()

    @staticmethod
    def synchronize(migrate=None):
        if migrate:
            _migrate_db_if_needed()

        _insert_new_workers_in_database()

    @staticmethod
    def create_admin(email):
        _create_admin_user(email)

    @staticmethod
    def start_scheduler(is_background_process=False):
        pid = _start_new_scheduler(is_background_process)
        print(f"Created a new scheduler with PID {pid}.")
        return pid


def main():
    fire.Fire(CLICommands)


if __name__ == "__main__":
    with app.app_context():
        main()



from flask_admin import Admin

from skidward.app import app
from skidward.models import db, User, Role, Worker, Task, Job
from skidward.web.views import (
    UserAdmin,
    RoleAdmin,
    JobView,
    TaskView,
    WorkerView,
    RunView,
    TaskConfigure,
    SkidwardView,
    LogoutMenuLink,
    LoginMenuLink,
)


# Initializing admin with flask app, name and template type
admin = Admin(
    app, name="Skidward-Admin", template_mode="bootstrap3", index_view=SkidwardView()
)

# Adding Models as Views to admin
admin.add_view(UserAdmin(User, db.session))
admin.add_view(RoleAdmin(Role, db.session))
admin.add_view(WorkerView(Worker, db.session))
admin.add_view(TaskView(Task, db.session))
admin.add_view(JobView(Job, db.session))
admin.add_view(RunView(name="Run", endpoint="run_task", url="/admin/task"))
admin.add_view(
    TaskConfigure(name="Configure", endpoint="add_context", url="/admin/task")
)
admin.add_view(SkidwardView(name="Skidward-Home", endpoint="index", url="/"))
admin.add_link(LoginMenuLink(name="Login", endpoint="login", url="/login"))
admin.add_link(LogoutMenuLink(name="Logout", endpoint="logout", url="/logout"))


if __name__ == "__main__":
    app.run()
